<?php

// Prevent any direct access to this file
defined('AUTH_ACCESS') or die('No direct script access.');

/**
 * Manage Ongage API http requests
 *
 * Class Ongage
 */
class OngageConnect {

	public function __construct()
	{
		// Load the configs from the config file
		$config                     = require_once 'Config.php';
		$this->visible_list_fields  = self::array_get($config, 'visible_list_fields'  , []);
		$this->editable_list_fields = self::array_get($config, 'editable_list_fields' , []);
		$credentials                = self::array_get($config, 'credentials'          , []);
		$expected_credentials       = [ 'USERNAME', 'PASSWORD', 'ACCOUNT_CODE' ];

		// Validate expected credentials were given
		foreach ($expected_credentials as $expected_credential)
		{
			// Validate all credentials are provided
			if ( ! self::array_get($credentials, $expected_credential))
				throw new Ongage_SDK_error(
						'Invalid credentials: {expected_credential} is missing, please add it in your Config.php file',
						[ '{expected_credential}' => $expected_credential ]

				);

			// Set the credential
			$this->{$expected_credential} = $credentials[$expected_credential];
		}
	}

	// Credentials - this should be changed with the config file only
	protected $USERNAME;
	protected $PASSWORD;
	protected $ACCOUNT_CODE;


	// Form helpers variables - this should be changed with the config file only
	protected $visible_list_fields;
	protected $editable_list_fields;

	// Base API URL - you should not change this ever
	const BASE_API_URL = 'https://api.ongage.net/';

	// Time formats stored in Ongage and thier assoc php date format
	// used for converting timestamp into human readable date
	public static $ongage_to_php_time_formats = [
		'mm/dd/yyyy'            => 'm/d/Y',
		'dd/mm/yyyy'            => 'd/m/Y',
		'yyyy/mm/dd'            => 'Y/m/d',
		'dd-mm-yyyy'            => 'd-m-Y',
		'mm-dd-yyyy'            => 'm-d-Y',
		'yyyy-mm-dd'            => 'Y-m-d',
		'dd/mm/yyyy hh24:mi'    => 'd/m/Y H:i',
		'mm/dd/yyyy hh24:mi'    => 'm/d/Y H:i',
		'yyyy/mm/dd hh24:mi'    => 'Y/m/d H:i',
		'dd-mm-yyyy hh24:mi'    => 'd-m-Y H:i',
		'mm-dd-yyyy hh24:mi'    => 'm-d-Y H:i',
		'yyyy-mm-dd hh24:mi'    => 'Y-m-d H:i',
		'dd/mm/yyyy hh24:mi:ss' => 'd/m/Y H:i:s',
		'mm/dd/yyyy hh24:mi:ss' => 'm/d/Y H:i:s',
		'yyyy/mm/dd hh24:mi:ss' => 'Y/m/d H:i:s',
		'dd-mm-yyyy hh24:mi:ss' => 'd-m-Y H:i:s',
		'mm-dd-yyyy hh24:mi:ss' => 'm-d-Y H:i:s',
		'yyyy-mm-dd hh24:mi:ss' => 'Y-m-d H:i:s',
	];

	/**
	 * Array of all possible cURL error types
	 * @var array
	 */
	protected $curl_error_types = [
		1  => 'CURLE_UNSUPPORTED_PROTOCOL',
		2  => 'CURLE_FAILED_INIT',
		3  => 'CURLE_URL_MALFORMAT',
		4  => 'CURLE_URL_MALFORMAT_USER',
		5  => 'CURLE_COULDNT_RESOLVE_PROXY',
		6  => 'CURLE_COULDNT_RESOLVE_HOST',
		7  => 'CURLE_COULDNT_CONNECT',
		8  => 'CURLE_FTP_WEIRD_SERVER_REPLY',
		9  => 'CURLE_REMOTE_ACCESS_DENIED',
		11 => 'CURLE_FTP_WEIRD_PASS_REPLY',
		13 => 'CURLE_FTP_WEIRD_PASV_REPLY',
		14 => 'CURLE_FTP_WEIRD_227_FORMAT',
		15 => 'CURLE_FTP_CANT_GET_HOST',
		17 => 'CURLE_FTP_COULDNT_SET_TYPE',
		18 => 'CURLE_PARTIAL_FILE',
		19 => 'CURLE_FTP_COULDNT_RETR_FILE',
		21 => 'CURLE_QUOTE_ERROR',
		22 => 'CURLE_HTTP_RETURNED_ERROR',
		23 => 'CURLE_WRITE_ERROR',
		25 => 'CURLE_UPLOAD_FAILED',
		26 => 'CURLE_READ_ERROR',
		27 => 'CURLE_OUT_OF_MEMORY',
		28 => 'CURLE_OPERATION_TIMEDOUT',
		30 => 'CURLE_FTP_PORT_FAILED',
		31 => 'CURLE_FTP_COULDNT_USE_REST',
		33 => 'CURLE_RANGE_ERROR',
		34 => 'CURLE_HTTP_POST_ERROR',
		35 => 'CURLE_SSL_CONNECT_ERROR',
		36 => 'CURLE_BAD_DOWNLOAD_RESUME',
		37 => 'CURLE_FILE_COULDNT_READ_FILE',
		38 => 'CURLE_LDAP_CANNOT_BIND',
		39 => 'CURLE_LDAP_SEARCH_FAILED',
		41 => 'CURLE_FUNCTION_NOT_FOUND',
		42 => 'CURLE_ABORTED_BY_CALLBACK',
		43 => 'CURLE_BAD_FUNCTION_ARGUMENT',
		45 => 'CURLE_INTERFACE_FAILED',
		47 => 'CURLE_TOO_MANY_REDIRECTS',
		48 => 'CURLE_UNKNOWN_TELNET_OPTION',
		49 => 'CURLE_TELNET_OPTION_SYNTAX',
		51 => 'CURLE_PEER_FAILED_VERIFICATION',
		52 => 'CURLE_GOT_NOTHING',
		53 => 'CURLE_SSL_ENGINE_NOTFOUND',
		54 => 'CURLE_SSL_ENGINE_SETFAILED',
		55 => 'CURLE_SEND_ERROR',
		56 => 'CURLE_RECV_ERROR',
		58 => 'CURLE_SSL_CERTPROBLEM',
		59 => 'CURLE_SSL_CIPHER',
		60 => 'CURLE_SSL_CACERT',
		61 => 'CURLE_BAD_CONTENT_ENCODING',
		62 => 'CURLE_LDAP_INVALID_URL',
		63 => 'CURLE_FILESIZE_EXCEEDED',
		64 => 'CURLE_USE_SSL_FAILED',
		65 => 'CURLE_SEND_FAIL_REWIND',
		66 => 'CURLE_SSL_ENGINE_INITFAILED',
		67 => 'CURLE_LOGIN_DENIED',
		68 => 'CURLE_TFTP_NOTFOUND',
		69 => 'CURLE_TFTP_PERM',
		70 => 'CURLE_REMOTE_DISK_FULL',
		71 => 'CURLE_TFTP_ILLEGAL',
		72 => 'CURLE_TFTP_UNKNOWNID',
		73 => 'CURLE_REMOTE_FILE_EXISTS',
		74 => 'CURLE_TFTP_NOSUCHUSER',
		75 => 'CURLE_CONV_FAILED',
		76 => 'CURLE_CONV_REQD',
		77 => 'CURLE_SSL_CACERT_BADFILE',
		78 => 'CURLE_REMOTE_FILE_NOT_FOUND',
		79 => 'CURLE_SSH',
		80 => 'CURLE_SSL_SHUTDOWN_FAILED',
		81 => 'CURLE_AGAIN',
		82 => 'CURLE_SSL_CRL_BADFILE',
		83 => 'CURLE_SSL_ISSUER_ERROR',
		84 => 'CURLE_FTP_PRET_FAILED',
		85 => 'CURLE_RTSP_CSEQ_ERROR',
		86 => 'CURLE_RTSP_SESSION_ERROR',
		87 => 'CURLE_FTP_BAD_FILE_LIST',
		88 => 'CURLE_CHUNK_FAILED'
	];

	//-- HTTP API request to Ongage methods --------------------------------------------/

	/**
	 * Get your contact by its email and list_id
	 *
	 * @param  string    $email
	 * @param  int       $list_id
	 * @param  string    $contact_id
	 * @return mixed
	 * @throws Exception
	 */
	public function get_contact($email, $list_id, $contact_id)
	{
		// Prepare request data
		$contact_request_payload = [
			'list_id'  => $list_id,
		];

		// Send the request and retrieve the contact
		$contact = $this->execute_api_call(
			'GET',
			'api/contacts/by_email/'.$email,
			$contact_request_payload
		);

		$contact_payload    = self::array_get($contact, 'payload');
		$payload_contact_id = self::array_get($contact_payload, 'id');

		// Validate contact id matches the provided one
		if ($payload_contact_id != $contact_id)
			throw new Ongage_SDK_error('Invalid contact id was given',[], 403);

		return $contact_payload;
	}

	/**
	 * Get all your list fields
	 *
	 *
	 * @param  int       $list_id
	 * @return array
	 * @throws Exception
	 */
	public function get_list_fields($list_id, $index_by_names = TRUE)
	{
		$list_field_data = $this->execute_api_call(
			'GET',
			'api/list_fields',
			[
				'list_id' => $list_id
			]
		);

		$list_fields_payload = self::array_get($list_field_data, 'payload');

		// List fields will be returned when they are indexed by their names as default
		return $index_by_names
			? array_column($list_fields_payload, NULL, 'name')
			: $list_fields_payload;
	}

	/**
	 * Update contact
	 *
	 * @param  string $email
	 * @param  int    $list_id
	 * @param  array  $update_payload || array of [key => value ] of your contact's updated list fields
	 * @return array  $updated_contact
	 */
	public function update_contact($email, $list_id, array $update_payload)
	{
		$update_contact_payload = [
			'email'   => $email,
			'list_id' => $list_id,
			'fields'  => $update_payload
		];

		try
		{
			$updated_contact = $this->execute_api_call(
				'PUT',
				'api/v2/contacts',
				$update_contact_payload
			);

			$payload = self::array_get($updated_contact, 'payload');

			return $payload;
		}
		catch (Ongage_Api_Error $e)
		{
			// Extract the original payload to get the correct error messages for this contact
			$error_payload     = $e->get_response_payload();
			$raw_update_errors = self::array_get($error_payload, 'failed_emails', []);
			$raw_update_errors = self::array_get($raw_update_errors, $email);


			// Prepare the errors and throw the exception
			$prepared_update_errors = [];
			foreach ($raw_update_errors as $field => $error_message)
			{
				$prepared_update_errors[] = $field.': '.$error_message;
			}

			throw new Ongage_Api_Error(
				'You have failed to update your preferences',
				[],
				$e->getCode(),
				$prepared_update_errors
			);

		}
	}

	/**
	 * Changes contact status to unsubscribe
	 *
	 * @param string  $email
	 * @param integer $list_id
	 * @return array           || returns unsubscribed contact
	 */
	public function unsubscribe($email, $list_id)
	{
		return $this->change_status($email, $list_id,'unsubscribe');
	}

	/**
	 * Changes contact status to resubscribe
	 *
	 * @param string  $email
	 * @param integer $list_id
	 * @return array           || returns resubscribe contact
	 */
	public function resubscribe($email, $list_id)
	{
		return $this->change_status($email, $list_id, 'resubscribe');
	}

	/**
	 * Hard deletes a contact
	 *
	 * @param string  $contact_id
	 * @param integer $list_id
	 * @return array              || deleted contact
	 * @throws Exception
	 */
	public function delete_contact($contact_id, $list_id)
	{
		$delete_payload = [
			'contact_id' => $contact_id,
			'list_id'    => $list_id
		];

		$deleted_contact = $this->execute_api_call(
			'POST',
			'api/contacts/delete',
			$delete_payload
		);

		$payload = self::array_get($deleted_contact, 'payload');
		return $payload;
	}

	/**
	 * Makes a change status request
	 *
	 * @param string  $email
	 * @param integer $list_id
	 * @param string  $new_status
	 * @return array               || returns status-changed contact
	 * @throws Exception
	 */
	protected function change_status($email, $list_id, $new_status)
	{
		$request_payload = [
			'list_id'   => $list_id,
			'change_to' => $new_status,
			'emails'    => [ $email ]
		];

		$updated_contact = $this->execute_api_call(
			'POST',
			'api/v2/contacts/change_status',
			$request_payload
		);

		$payload = self::array_get($updated_contact, 'payload');

		return $payload;
	}

	/**
	 * Executes API request to Ongage
	 *
	 * @param  string     $request_method    || POST / PUT / GET ...
	 * @param  string     $api_url_end_point || api/contacts/by_email/test@test.com
	 * @param  array      $request_payload   || array of the payload data - as defined in the http://apidocs.ongage.net for each request
	 * @return mixed
	 * @throws Exception
	 */
	protected function execute_api_call($request_method, $api_url_end_point, array $request_payload = [])
	{

		$url_full_end_point = self::BASE_API_URL.$api_url_end_point;

		// Convert request payload according to the request method type
		if ($request_payload)
		{
			if ($request_method === 'GET')
			{
				$url_full_end_point.= '?' . http_build_query($request_payload);
			}
			else
			{
				$request_payload = json_encode($request_payload);
			}
		}

		// Set request headers
		$request_headers = [
			// Credentials
			'X_USERNAME: '.$this->USERNAME,
			'X_PASSWORD: '.$this->PASSWORD,
			'X_ACCOUNT_CODE: '.$this->ACCOUNT_CODE
		];

		if ($request_method !== 'GET')
		{
			$request_headers = array_merge(
				$request_headers,
				[
					// Let the API know you are about to send JSON
					'Content-Type: application/json; charset=utf-8',
					'Content-Length: '.strlen($request_payload),
					'Accept: application/json',
				]
			);
		}


		// Start CURL HTTP request
		$ch = curl_init();
		// Set request URL
		curl_setopt($ch, CURLOPT_URL, $url_full_end_point);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, TRUE);


		// Set HTTP request type
		switch ($request_method)
		{
			case 'POST':
				curl_setopt($ch, CURLOPT_POST, TRUE);
				break;
			case 'PUT':
				curl_setopt($ch, CURLOPT_CUSTOMREQUEST, 'PUT');
				break;
		}

		// Set the request headers and body
		if ($request_method !== 'GET')
		{
			curl_setopt($ch, CURLOPT_POSTFIELDS, $request_payload);
		}
		curl_setopt($ch, CURLOPT_HTTPHEADER, $request_headers);
		curl_setopt($ch, CURLINFO_CONTENT_TYPE, 'application/json');

		// Execute the request and get response code and response body
		$data               = curl_exec($ch);
		$httpcode           = curl_getinfo($ch, CURLINFO_HTTP_CODE);

		// Catch any cURL errors
		$curl_error_message = curl_error($ch);
		$curl_error_code    = curl_errno($ch);

		// Close the HTTP connection
		curl_close($ch);

		//-- Validations ---------------------------------------------------------------/

		// cURL error handling
		if ($curl_error_message)
		{
			$curl_error_type = self::array_get($this->curl_error_types, $curl_error_code);

			throw new Ongage_SDK_error(
				'cURL error "{curl_error_type}" ({curl_error_code}) with the following message: "{curl_error_message}"',
				[
					'{curl_error_message}' => $curl_error_message,
					'{curl_error_code}'    => $curl_error_code,
					'{curl_error_type}'    => $curl_error_type
				]
			);
		}

		if ( ! $data)
			throw new Ongage_Api_Error('No data was returned from API', [],500);

		// Convert data to PHP array (data will be returned as JSON)
		$data = json_decode($data, TRUE);

		// Just in case
		if ( ! is_array($data) )
			throw new Ongage_SDK_error(
				'Unexpected data was returned from API with the code: {httpcode}',
				[
					'{httpcode}' => $httpcode
				]

			);


		// If it is not a "HTTP 200 OK" - Validation error has failed in Ongage server
		if ($httpcode !== 200)
		{
			// Get the Ongage Validation error and throw an exception about it
			$error_payload = self::array_get($data, 'payload');

			throw new Ongage_Api_Error(
				self::array_get($error_payload, 'message'),
				[],
				$httpcode,
				self::array_get($error_payload, 'errors', []),
				$error_payload
			);
		}

		return $data;
	}

	//--- Helper functions -------------------------------------------------------------/

	/**
	 *
	 * Helper function for preparing your contacts reference form
	 *
	 * @param array $contact                  || contact payload as expected to come from GET api/contact/by_email/<email>
	 * @param array $list_fields              || list field payload with name indexes
	 * @return array $preference_form_inputs  || return an array of reference form
	 *
	 */
	public function prepare_preference_form_inputs(array $contact, array $list_fields)
	{
		$preference_form_inputs = [];

		// Separates $list_fields so it can be "glued" with the ocx_status in the correct position
		$email              =  array_slice($list_fields, 0, 1, TRUE);
		$rest_of_the_fields =  array_slice($list_fields, 1, count($list_fields) -1, TRUE);
		$status_arr         = [
			'ocx_status' => [
				'name'      => 'ocx_status',
				'title'     => 'Status',
				'mandatory' => FALSE,
				'type'      => 'system_field',
				'format'    => NULL
			]
		];

		// Glue the list field pieces back together
		$list_fields = $email + $status_arr + $rest_of_the_fields;

		// Loop the list field to create the inputs
		$always_shown    = ['email', 'ocx_status' ];
		$always_disabled = ['email', 'ocx_status' ];
		foreach ($list_fields as $list_field_name => $list_field)
		{
			// Ignore not allowed inputs so they wont be included
			$warnings = [];
			if ( ! in_array($list_field['name'], $always_shown) && $this->visible_list_fields && ! in_array($list_field['name'], $this->visible_list_fields))
			{
				continue;
			}

			// Set default value as defined in your Ongage list field
			$value        = self::array_get($contact, $list_field['name'])?: $list_field['default'];
			// Set place holder
			$place_holder = $list_field['format']
				? $list_field['title'].' ('.$list_field['format'].')'
				: $list_field['title'];

			// Mark any mandatory fields
			if ($list_field['mandatory'] && $list_field['name'] != 'email')
			{
				$warnings[] = [
					'type'    => 'warning',
					'message' => 'Mandatory'
				];
			}

			// In ongage we save all dates as UTC - you may want to convert this to your contact time zone
			if ($list_field['type'] == 'date' && $value)
			{
				$php_time_format = self::array_get(self::$ongage_to_php_time_formats, $list_field['format']);
				$value           = date($php_time_format, $value);
				$warnings[] = [
					'type'    => 'primary',
					'message' => 'UTC date'
				];
			}


			// Define if this field will disabled
			$disabled =
				in_array($list_field['name'], $always_disabled) ||
				(
					$this->editable_list_fields &&
					! in_array($list_field['name'], $this->editable_list_fields)
				);

			// Add the rest of the input values
			$preference_form_inputs[] = [
				'name'         => $list_field['name'],
				'disabled'     => $disabled,
				'title'        => $list_field['title'],
				'mandatory'    => $list_field['mandatory'] && $list_field['name'] != 'email',
				'value'        => $value,
				'warnings'     => $warnings,
				'place_holder' => $place_holder,
			];
		}

		return $preference_form_inputs;
	}

	//--- Helper functions -------------------------------------------------------------/

	/**
	 * Retrieve a single key from an array. If the key does not exist in the
	 * array, the default value will be returned instead.
	 *
	 *     // Get the value "email" from $_POST, if it exists
	 *     $email = Ongage::array_get($_POST, 'email');
	 *
	 *     // Get the value "list_id" from $_GET, if it exists
	 *     $list_id = Ongage::array_get($_GET, 'list_id');
	 *
	 * @param   array   $array      array to extract from
	 * @param   string  $key        key name
	 * @param   mixed   $default    default value
	 * @return  mixed
	 *
	 * IMPORTANT - This function is part of the kohana PHP framework
	 *
	 * @copyright  (c) 2007-2012 Kohana Team
	 * @license    http://kohanaframework.org/license
	 */
	public static function array_get($array, $key, $default = NULL)
	{
		return isset($array[$key]) ? $array[$key] : $default;
	}

	/**
	 * Retrieves multiple keys from an array (Only the keys that exist in the array)
	 *
	 *     // Get the the keys ('name' & 'title') from $data
	 *     $data = ['name' => 'Tony', 'age' => 25];
	 *     Ongage::array_extract_exist($data, ['name', 'title'];
	 *     // Result => ['name' => 'Tony'];
	 *
	 *     // Get the the keys ('name' & 'title') from $data (set default value for 'title' key)
	 *     $data = ['name' => 'Tony', 'age' => 25];
	 *     Ongage::array_extract_exist($data, ['name', 'title', [], ['title' => 'Hello']];
	 *     // Result => ['name' => 'Tony', 'title' => 'Hello'];
	 *
	 * @param   array  $array           array to extract keys from
	 * @param   array  $keys            list of keys
	 * @param   array  $boolean_keys    list of boolean keys
	 * @param   array  $default_values  default values in case key not exist
	 * @param   array  $array_keys      list of array keys (need to explode comma separated strings)
	 * @return  array
	 */
	public static function array_extract_exist($array, array $keys, array $boolean_keys = [], array $default_values = [], array $array_keys = [])
	{
		$found = [];
		foreach ($keys as $key)
		{
			if (array_key_exists($key, $array))
			{
				$found[$key] = $array[$key];
				$found[$key] === 'null'
				AND $found[$key] = NULL;
			}
			elseif (array_key_exists($key, $default_values))
			{
				$found[$key] = $default_values[$key];
			}
		}

		foreach ($boolean_keys as $key)
		{
			if (array_key_exists($key, $array))
			{
				$found[$key] = $array[$key] && $array[$key] !== 'false';
			}
			elseif (array_key_exists($key, $default_values))
			{
				$found[$key] = $default_values[$key];
			}
		}

		foreach ($array_keys as $key)
		{
			if (array_key_exists($key, $array))
			{
				if (is_array($array[$key]))
					$found[$key] = $array[$key];
				else
					$found[$key] = explode(',', $array[$key]);
			}
			elseif (array_key_exists($key, $default_values))
			{
				$found[$key] = $default_values[$key];
			}
		}

		return $found;
	}

	/**
	 * Prints put all arguments given amd kills the program
	 *    Ongage::debug_vars('hello', ['name', 'title'];
	 * 	  this will print the following:
	 *
	 * Array
	 * (
	 *     [0] => hello
	 *     [1] => Array
	 *         (
	 *             [0] => name
	 *             [1] => title
	 *         )
	 * )
	 *
	 *
	 */
	public static function debug_vars()
	{
		if (func_num_args() === 0)
			return;

		// Get all passed variables
		$variables = func_get_args();

		echo '<pre>';
		print_r($variables);
		echo '</pre>';
		die;
	}

}

/**
 * Handle all Ongage API errors
 *
 * Class Ongage_Api_Error
 */
class Ongage_Api_Error extends Exception {

	protected $code;
	protected $errors;
	protected $message;
	protected $response_payload;

	/**
	 * Ongage_Api_Error constructor.
	 *
	 * @param string $message   || The message that will be outputed
	 * @param array  $variables || Array of strings that can replace template string dynamically ||
	 * @param int    $code      || Http error code
	 * @param array  $errors    || Array of additional errors
	 */
	public function __construct($message = "", array $variables = [], $code = 412, array $errors = [], array $response_payload = [])
	{
		// Set the message
		$this->message = $variables ? strtr($message, $variables) : $message;

		$this->errors  = [
			'http_code'        => $code,
			'original_message' => $message,
			'errors'           => $errors
		];

		$this->code             = $code;
		$this->response_payload = $response_payload;

		// Pass the message and integer code to the parent
		parent::__construct($this->message, (int) $code);

	}

	/**
	 * Retrieve the errors array
	 *
	 * @return array
	 */
	public function get_errors()
	{
		return $this->errors;
	}

	/**
	 * Retrieve the original API error payload
	 *
	 * @return array
	 */
	public function get_response_payload()
	{
		return $this->response_payload;
	}
}

/**
 * Handle inner Ongage SDK errors
 *
 * Class Ongage_SDK_error
 */
class Ongage_SDK_error extends Exception
{

	/**
	 * Ongage_SDK_error constructor.
	 *
	 * @param string $message
	 * @param array  $variables
	 */
	public function __construct($message = "", array $variables = [])
	{
		// Set the message
		if ($variables)
		{
			$message = strtr($message, $variables);
		}

		// Pass the message and integer code to the parent
		parent::__construct($message);

	}
}
